#include <stdio.h>
#include <stdlib.h>
#include "loft_c.h"
#include "order.h"

void set_current_scheduler  (scheduler_t);

void internal_error         (char *format, ...);
void verify                 (scheduler_t,char*);
void empty_fields           (scheduler_t);

//#define VERIFY(sched,msg)  verify (sched,msg)
//#define EMPTYFIELDS(sched) empty_fields (sched)

#define VERIFY(sched,msg)
#define EMPTYFIELDS(sched)

int direct_access  = 1;
int trace_instants = 0;

/***************************************************/
static void add_to_orders (int kind,thread_t thread)
{
   if (!thread || thread->magic != MAGIC) return;
   _append_item_list (thread->scheduler->orders,make_order (kind,thread));
}

void add_to_link (scheduler_t sched,thread_t thread)
{
   add_to_orders (LINK_ORDER,thread);
}

void add_to_unlink (scheduler_t sched,thread_t thread)
{
   add_to_orders (UNLINK_ORDER,thread);
}

void add_to_broadcast (scheduler_t sched,broadcast_t order)
{
   _append_item_list (sched->to_broadcast,order);
}

/***************************************************/
void stop (thread_t thread)
{
   add_to_orders (STOP_ORDER,thread);
}
 
void suspend (thread_t thread)
{
   add_to_orders (SUSPEND_ORDER,thread);
}
 
void resume (thread_t thread)
{
   add_to_orders (RESUME_ORDER,thread);
}
 
void thread_deallocate (thread_t thread)
{
   add_to_orders (DEALLOCATE_ORDER,thread);
}
 
/***************************************************/
scheduler_t scheduler_create (void)
{
   scheduler_t sched = mymalloc (sizeof (struct _scheduler_t));
   sched->linked          = _make_item_list ();
   sched->orders          = _make_item_list ();
   sched->to_broadcast    = _make_item_list ();
   sched->to_finalize     = _make_item_list ();
   sched->timer           = _make_item_list ();
   sched->move            = 0;
   sched->eoi             = 0;
   sched->instant         = 0;
   sched->active          = 0;
   sched->determinist     = 1;   
   return sched;
}

scheduler_t scheduler_nd_create (void)
{
   scheduler_t sched = scheduler_create ();
   sched->determinist = 0;   
   return sched;
}

void scheduler_deallocate (scheduler_t sched)
{
   _deallocate_item_list (sched->linked);
   _deallocate_item_list (sched->orders);
   _deallocate_item_list (sched->to_broadcast);
   _deallocate_item_list (sched->to_finalize);
   _deallocate_item_list (sched->timer);      
   myfree (sched);
}

/****************************************************/
// Phase
/****************************************************/
static int fire_one (item_cell_t cell,void *arg)
{
   thread_t thread = _content_of_cell (cell);
   scheduler_t sched = arg;
   if (fireable (thread)) {
      sched->current_cell = cell;
      fire_thread (thread);
      if (thread->state != _CONT)  sched->active--;
      if (direct_access && thread->state == _WAIT) return MASK;
   }
   return KEEP;
}

static void fire_all_threads (scheduler_t sched)
{
   VERIFY (sched,"before");
   _apply_on_cells (sched->linked,fire_one,sched);
   VERIFY (sched,"after");   
}

/****************************************************/
// Transfer Events
/****************************************************/
static int transfer_event (void *item,void *arg)
{
   broadcast_t order = item;
   if (order->pure) broadcast (order->event);
   else broadcast_value (order->event,order->value);
   free (order);
   return REMOVE;
}

static void transfer_all_events (scheduler_t sched)
{
   _apply_on_items (sched->to_broadcast,transfer_event,NULL);
}

/****************************************************/
// Processing of orders
/****************************************************/
static int process_order (void *item,void *arg)
{
   order_t order = item;
   scheduler_t sched = arg;
   thread_t thread = order->thread;
   switch (order->kind) {
      case LINK_ORDER:
	 thread->scheduler = sched;
	 _append_item_list (sched->linked,thread);
	 break;
      case UNLINK_ORDER:
	 thread->scheduler = NULL;
	 break;
      case SUSPEND_ORDER:
	 thread->suspended = 1;
	 break;
      case RESUME_ORDER:
	 thread->suspended = 0;
	 break;
      case STOP_ORDER:
	 if (thread->state != _TERM) {
	    _append_item_list (sched->to_finalize,thread);
	 }
	 thread->state = _TERM;
	 break;
      case DEALLOCATE_ORDER:
	 deallocate_thread (thread);
	 break;
   }
   free (order);
   return REMOVE;
}

static void process_all_orders (scheduler_t sched)
{
   _apply_on_items (sched->orders,process_order,sched);
}

/****************************************************/
static int reset_thread (void *item,void *arg)
{
   thread_t thread = item;
   if (_TERM == thread->state || NULL == thread->scheduler)
      return REMOVE;
   if (reset_thread_state (thread)) thread->scheduler->active++;
   return KEEP;
}

static void reset_all_threads (scheduler_t sched)
{
   sched->active = 0;
   _apply_on_items (sched->linked,reset_thread,NULL);
}

/****************************************************/
// Timer
/****************************************************/
void register_in_timer (scheduler_t sched)
{
   item_cell_t cell =
      _append_item_list (sched->timer,sched->current_cell);
   self ()->timer_cell = cell;   
}

static int process_deadline (void *item,void *arg)
{
   item_cell_t cell = item;
   thread_t thread = _content_of_cell (cell);
   scheduler_t sched = arg;
   if (thread->deadline > sched->instant) return KEEP;
   if (thread->state == _WAIT) {
      unmask (sched,cell);
      unregister_in_event (thread);
   }
   return REMOVE;
}

static void process_all_deadlines (scheduler_t sched)
{
   _apply_on_items (sched->timer,process_deadline,sched);
}

/****************************************************/
// Scheduler Reset
/****************************************************/
static void reset_scheduler (scheduler_t sched)
{
   sched->move = sched->eoi = 0;
   sched->instant++;
   process_all_deadlines (sched);
}

/*****************************************************/
// finalization of threads (when stopped)
/*****************************************************/
static int finalize_one_thread (void *item,void *arg)
{
   finalize_thread ((thread_t)item);
   return REMOVE;
}

static void finalize_stopped_threads (scheduler_t sched)
{
   _apply_on_items (sched->to_finalize,finalize_one_thread,NULL);
}

/****************************************************/
// Start Instant
/****************************************************/
static void start_instant (scheduler_t sched)
{
   reset_scheduler          (sched);
   process_all_orders       (sched);
   transfer_all_events      (sched);
   reset_all_threads        (sched);
   finalize_stopped_threads (sched);
   
   VERIFY (sched,"start");
   EMPTYFIELDS (sched);
}

/****************************************************/
// Instant
/****************************************************/
static void make_decision (scheduler_t sched)
{
   if (sched->move) sched->move = 0; else sched->eoi = 1;
}

static int stable (scheduler_t sched)
{
   return sched->active == 0;
}

void instant (scheduler_t sched)
{
   if (trace_instants) {
      printf ("\ninstant %d (%d): ",sched->instant,
	      _length_of_item_list (sched->linked));
      fflush (stdout);
   }
   start_instant (sched);
   while (1) {
      fire_all_threads (sched);
      make_decision (sched);
      if (stable (sched)) return;
   }
}

/****************************************************/
void scheduler_instant (scheduler_t sched)
{
   set_current_scheduler (sched);
   instant (sched);
}

/****************************************************/
void scheduler_start (scheduler_t sched)
{
   while (1) scheduler_instant (sched);
}

/*****************************************************/
// unmasking
/*****************************************************/
void unmask (scheduler_t sched,item_cell_t cell)
{
   thread_t thread = _content_of_cell (cell);
   if (direct_access)
      _reinsert_cell (sched->linked,cell,sched->determinist);
   if (thread->suspended) return;
   thread->state =_CONT;
   sched->active++;
}

/****************************************************/
void link_to_scheduler (scheduler_t sched,thread_t thread)
{
   add_to_link (sched,thread);
}

/****************************************************/
/****************************************************/
void verify (scheduler_t sched,char *s)
{
   int n1 = sched->active;
   int n2 = 0;
   item_cell_t cell = _first_cell (sched->linked);
   while (cell) {
      thread_t thread = _content_of_cell (cell);
      if (!thread->suspended && thread->state == _CONT) n2++;
      cell = _next_cell (cell);
   }
   if (n1 != n2)
      internal_error ("%d,%d) - %s !!!!!!!!!!!\n",n1,n2,s);
}

void empty_fields (scheduler_t sched)
{
   if (_length_of_item_list (sched->orders))
      internal_error ("non empty order list %d\n",
		      _length_of_item_list (sched->orders));

   if (_length_of_item_list (sched->to_broadcast))
      internal_error ("non empty broadcast list\n");
}
