#include <stdio.h>
#include <stdlib.h>
#include "status.h"
#include "loft_sem.h"
#include "order.h"

void set_current_scheduler (scheduler_t);
int trace_instants = 0;

/***************************************************/
static void add_to_orders (int kind,thread_t thread)
{
   _append_item_list (thread->scheduler->orders,make_order (kind,thread));
}

void add_to_link (scheduler_t sched,thread_t thread)
{
   add_to_orders (LINK_ORDER,thread);
}

void add_to_unlink (thread_t thread)
{
   add_to_orders (UNLINK_ORDER,thread);
}

void add_to_broadcast (scheduler_t sched,broadcast_t order)
{
   _append_item_list (sched->to_broadcast,order);
}

/**************************************************/
void stop (thread_t thread)
{
   add_to_orders (STOP_ORDER,thread);
}
 
void suspend (thread_t thread)
{
   add_to_orders (SUSPEND_ORDER,thread);
}
 
void resume (thread_t thread)
{
   add_to_orders (RESUME_ORDER,thread);
}

/***************************************************/
scheduler_t scheduler_create (void)
{
   scheduler_t sched = mymalloc (sizeof (struct _scheduler_t));
   sched->linked          = _make_item_list ();
   sched->orders          = _make_item_list ();
   sched->to_broadcast    = _make_item_list ();
   sched->to_finalize     = _make_item_list ();
   sched->move            = 0;
   sched->eoi             = 0;
   sched->events          = _make_item_list ();
#ifdef OPTIM
   sched->timer           = _make_item_list ();
#endif
   return sched;
}

/****************************************************/
// Phase
/****************************************************/
static int fire_one (void *item,void *arg)
{
   thread_t thread = item;
   if (fireable (thread)) fire_thread (thread);
   return KEEP;
}

static void fire_all_threads (scheduler_t sched)
{
    _apply_on_items (sched->linked,fire_one,NULL);
}

/****************************************************/
// Transfer Events
/****************************************************/
static int transfer_event (void *item,void *arg)
{
   broadcast_t order = item;
   if (order->pure) broadcast (order->event);
   else broadcast_value (order->event,order->value);
   free (order);
   return REMOVE;
}

static void transfer_all_events (scheduler_t sched)
{
   _apply_on_items (sched->to_broadcast,transfer_event,NULL);
}

/****************************************************/
// Processing of orders
/****************************************************/
static int process_order (void *item,void *arg)
{
   order_t order = item;
   scheduler_t sched = arg;
   thread_t thread = order->thread;
   switch (order->kind) {
      case LINK_ORDER:
	 thread->scheduler = sched;
	 _append_item_list (sched->linked,thread);
	 break;
      case UNLINK_ORDER:
	 thread->scheduler = NULL;
	 break;
      case SUSPEND_ORDER:
	 thread->suspended = 1;
	 break;
      case RESUME_ORDER:
	 thread->suspended = 0;
	 break;
      case STOP_ORDER:
	 if (thread->state != _TERM) {
	    _append_item_list (sched->to_finalize,thread);
	 }
	 thread->state = _TERM;
	 break;
   }
   free (order);
   return REMOVE;
}

static void process_all_orders (scheduler_t sched)
{
   _apply_on_items (sched->orders,process_order,sched);
}

/****************************************************/
static int reset_thread (void *item,void *arg)
{
   thread_t thread = item;
   if (_TERM == thread->state || NULL == thread->scheduler)
      return REMOVE;
   reset_thread_state (thread);
   return KEEP;
}

static void reset_all_threads (scheduler_t sched)
{
   _apply_on_items (sched->linked,reset_thread,NULL);
}

/****************************************************/
// Timer
/****************************************************/
#ifdef OPTIM
void register_in_timer (scheduler_t sched)
{
   _append_item_list (sched->timer,self ());
}

static int process_deadline (void *item,void *arg)
{
   thread_t thread = item;
   thread->deadline--;  
   if (thread->deadline > 0) return KEEP;
   if (thread->state == _WAIT) thread->state = _CONT;
   return REMOVE;
}

static void process_all_deadlines (scheduler_t sched)
{
   _apply_on_items (sched->timer,process_deadline,sched);
}
#endif

/****************************************************/
// Scheduler Reset
/****************************************************/
static void reset_scheduler (scheduler_t sched)
{
   sched->move = sched->eoi = 0;
   _reset_item_list (sched->events);
#ifdef OPTIM
   process_all_deadlines (sched);
#endif
}

/*****************************************************/
// finalization of threads (when stopped)
/*****************************************************/
static int finalize_one_thread (void *item,void *arg)
{
   finalize ((thread_t)item);
   return REMOVE;
}

static void finalize_stopped_threads (scheduler_t sched)
{
   _apply_on_items (sched->to_finalize,finalize_one_thread,NULL);
}

/****************************************************/
// Start Instant
/****************************************************/
static void start_instant (scheduler_t sched)
{
   reset_scheduler          (sched);
   process_all_orders       (sched);
   transfer_all_events      (sched);
   reset_all_threads        (sched);
   finalize_stopped_threads (sched);
}

/****************************************************/
// Instant
/****************************************************/
static void make_decision (scheduler_t sched)
{
   if (sched->move) sched->move = 0; else sched->eoi = 1;
}

static int stable (scheduler_t sched)
{
   item_cell_t cell = _first_cell (sched->linked);
   while (cell) {
      if (fireable ((thread_t)_content_of_cell (cell))) return 0;
      cell = _next_cell (cell);
   }
   return 1;
}

void instant (scheduler_t sched)
{
   if (trace_instants) {
      static int i = 0;
      printf ("\ninstant %d (%d): ",i++,
	      _length_of_item_list (sched->linked));
   }
   start_instant (sched);
   while (1) {
      fire_all_threads (sched);
      make_decision (sched);
      if (stable (sched)) return;
   }
}

/****************************************************/
void scheduler_start (scheduler_t sched)
{
   while (1) scheduler_instant (sched);
}

/****************************************************/
void scheduler_instant (scheduler_t sched)
{
   set_current_scheduler (sched);
   instant (sched);
}


